/*************************************************** */
/* Rule Set Based Access Control                     */
/* Author and (c) 1999-2009: Amon Ott <ao@rsbac.org> */
/* (a lot copied from mm/slab.c, with other          */
/*  copyrights)                                      */
/* Memory allocation functions for all parts         */
/* Last modified: 30/Mar/2009                        */
/*************************************************** */

#include <rsbac/types.h>
#include <rsbac/rkmem.h>
#include <rsbac/debug.h>
#include <rsbac/aci.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <linux/timer.h>

/* Size description struct for general RSBAC caches. */
typedef struct rsbac_cache_sizes {
	size_t		 cs_size;
	struct kmem_cache *cs_cachep;
	char             name[15];
} rsbac_cache_sizes_t;

static rsbac_cache_sizes_t rsbac_cache_sizes[] = {
#if PAGE_SIZE == 4096
	{    32,	NULL,	"rsbac-32"},
#endif
	{    64,	NULL,	"rsbac-64"},
#if PAGE_SIZE == 4096
	{    96,	NULL,	"rsbac-96"},
#endif
	{   128,	NULL,	"rsbac-128"},
	{   192,	NULL,	"rsbac-192"},
	{   256,	NULL,	"rsbac-256"},
	{   384,	NULL,	"rsbac-384"},
	{   512,	NULL,	"rsbac-512"},
	{  1024,	NULL,	"rsbac-1024"},
	{  2048,	NULL,	"rsbac-2048"},
	{  4096,	NULL,	"rsbac-4096"},
	{  8192,	NULL,	"rsbac-8192"},
	{ 16384,	NULL,	"rsbac-16384"},
	{ 32768,	NULL,	"rsbac-32768"},
	{ 65536,	NULL,	"rsbac-65536"},
	{131072,	NULL,	"rsbac-131072"},
#ifndef CONFIG_MMU
        {262144,        NULL,   "rsbac-262144"},
        {524288,        NULL,   "rsbac-524288"},
        {1048576,        NULL,   "rsbac-1048576"},
#ifdef CONFIG_LARGE_ALLOCS
        {2097152,        NULL,   "rsbac-2097152"},
        {4194304,        NULL,   "rsbac-4194304"},
        {8388608,        NULL,   "rsbac-8388608"},
        {16777216,        NULL,   "rsbac-16777216"},
        {33554432,        NULL,   "rsbac-33554432"},
#endif /* CONFIG_LARGE_ALLOCS */
#endif /* CONFIG_MMU */
	{     0,	NULL,	"rsbac0"}
};
#ifdef CONFIG_RSBAC_INIT_DELAY
void rsbac_kmem_cache_sizes_init(void)
#else
void __init rsbac_kmem_cache_sizes_init(void)
#endif
{
	rsbac_cache_sizes_t *sizes = rsbac_cache_sizes;

	while (sizes->cs_size) {
		sizes->cs_cachep = kmem_cache_create(
			sizes->name, sizes->cs_size,
			0, SLAB_HWCACHE_ALIGN|SLAB_DESTROY_BY_RCU, NULL);
		if (!sizes->cs_cachep)
			BUG();

		sizes++;
	}
}


/**
 * rsbac_kmalloc - allocate memory
 * @size: how many bytes of memory are required.
 *
 * rsbac_kmalloc is the normal method of allocating memory for RSBAC
 * in the kernel. It will always be of type GFP_KERNEL in 2.4 and
 * GFP_ATOMIC in 2.6.
 *
 * rsbac_kmalloc'd memory is freed by rsbac_kfree
 */
#if defined(CONFIG_RSBAC_REG) || defined(CONFIG_RSBAC_REG_MAINT)
EXPORT_SYMBOL(rsbac_kmalloc);
#endif
void * rsbac_kmalloc (size_t size)
{
	rsbac_cache_sizes_t *csizep = rsbac_cache_sizes;

        if(!size)
          return NULL;
        if(!csizep->cs_cachep)
          return kmalloc(size, GFP_ATOMIC);
	for (; csizep->cs_size; csizep++) {
		if (size > csizep->cs_size)
			continue;
		return kmem_cache_alloc(csizep->cs_cachep, GFP_ATOMIC);
	}
	rsbac_printk(KERN_WARNING
	       "rsbac_kmalloc: size %u requested, max size is %u!\n",
	       size, RSBAC_MAX_KMALLOC);
	BUG();
	return NULL;
}

#if defined(CONFIG_RSBAC_REG) || defined(CONFIG_RSBAC_REG_MAINT)
EXPORT_SYMBOL(rsbac_kmalloc_unlocked);
#endif
void * rsbac_kmalloc_unlocked (size_t size)
{
	rsbac_cache_sizes_t *csizep = rsbac_cache_sizes;

        if(!size)
          return NULL;
        if(!csizep->cs_cachep)
          return kmalloc(size, GFP_KERNEL);
	for (; csizep->cs_size; csizep++) {
		if (size > csizep->cs_size)
			continue;
		return kmem_cache_alloc(csizep->cs_cachep, GFP_KERNEL);
	}
	rsbac_printk(KERN_WARNING
	       "rsbac_kmalloc_unlocked: size %u requested, max size is %u!\n",
	       size, RSBAC_MAX_KMALLOC);
	BUG();
	return NULL;
}

#if defined(CONFIG_RSBAC_REG) || defined(CONFIG_RSBAC_REG_MAINT)
EXPORT_SYMBOL(rsbac_vkmalloc);
#endif
void * rsbac_vkmalloc (size_t size, rsbac_boolean_t * vmalloc_used_p)
{
	rsbac_cache_sizes_t *csizep = rsbac_cache_sizes;
	void * result;

        if(!size)
          return NULL;
	for (; csizep->cs_size; csizep++) {
		if (size > csizep->cs_size)
			continue;
		result = kmem_cache_alloc(csizep->cs_cachep, GFP_ATOMIC);
		if(result)
		  {
		    if(vmalloc_used_p)
		      *vmalloc_used_p = FALSE;
		    return result;
		  }
	}
        return NULL;
}

#if defined(CONFIG_RSBAC_REG) || defined(CONFIG_RSBAC_REG_MAINT)
EXPORT_SYMBOL(rsbac_vkmalloc_unlocked);
#endif
void * rsbac_vkmalloc_unlocked (size_t size, rsbac_boolean_t * vmalloc_used_p)
{
	rsbac_cache_sizes_t *csizep = rsbac_cache_sizes;
	void * result;

        if(!size)
          return NULL;
	for (; csizep->cs_size; csizep++) {
		if (size > csizep->cs_size)
			continue;
		result = kmem_cache_alloc(csizep->cs_cachep, GFP_KERNEL);
		if(result)
		  {
		    if(vmalloc_used_p)
		      *vmalloc_used_p = FALSE;
		    return result;
		  }
	}
        return NULL;
}

/* end of rkmem.c */
